import pytest
import os
import json
import time
from tenacity import retry, stop_after_attempt, wait_fixed
from requests.packages import urllib3

from python_sensor_sim.enums import SensorDeviceType
from psc_auth_integration.entities.auth_user_session import AuthUserSession
from assignments.common.utils import (
    sim_endpoint,
    create_asset_group,
    get_policies_summary,
    add_asset_group_to_endpoint,
    add_policy_override,
    setup_data,
    DEFENSE_URL,
    USERNAME,
    PASSWORD,
    ENV,
    LOCAL_PAS_HOST,
    EXPECTED_CALL_TO_SUCCEED,
    ASSET_GROUPS,
    ENDPOINTS
)
from assignments.tests.test_policy_preview import search_assets

urllib3.disable_warnings()

SESSION = None
POLICIES = None
CREATED_GROUPS = None


@pytest.fixture(scope="module", autouse=True)
def init():
    global SESSION, POLICIES
    os.environ["ENV"] = ENV
    SESSION = AuthUserSession(
        gateway_url=DEFENSE_URL, username=USERNAME, password=PASSWORD
    )
    setup_data(SESSION)
    POLICIES = get_policies_summary(SESSION).get("policies")


@pytest.mark.validate
def test_get_effective_policy_summary_group_with_no_policy_assigned():
    ag_group_1 = get_asset_group("effective-policy-test-1")
    ag_group_2 = get_asset_group("effective-policy-test-2")
    ag_group_5 = get_asset_group("effective-policy-test-5")

    response = get_effective_policy_summary(ag_group_5.get("id"))

    assert 4 == len(response["summaries"])

    total_assets = 0
    for summary in response["summaries"]:
        total_assets += summary["asset_count"]
    assert 6 == total_assets

    endpoint_1 = get_endpoint("pas-e2e-effective-1")
    endpoint_2 = get_endpoint("pas-e2e-effective-2")
    endpoint_3 = get_endpoint("pas-e2e-effective-3")
    endpoint_4 = get_endpoint("pas-e2e-effective-4")
    endpoint_5 = get_endpoint("pas-e2e-effective-5")
    endpoint_6 = get_endpoint("pas-e2e-effective-6")

    assert contains_summary_with(response["summaries"], 1, POLICIES[3].get("id"), [ag_group_1.get("id")], "GROUP", [ endpoint_3.get("id") ])
    assert contains_summary_with(response["summaries"], 2, POLICIES[3].get("id"), [ag_group_1.get("id"), ag_group_2.get("id")], "GROUP", [ endpoint_1.get("id"), endpoint_2.get("id") ])
    assert contains_summary_with(response["summaries"], 2, POLICIES[-1].get("id"), [], "MANUAL", [ endpoint_4.get("id"), endpoint_5.get("id") ])
    assert contains_summary_with(response["summaries"], 1, POLICIES[1].get("id"), [], "DEFAULT", [ endpoint_6.get("id") ])


@pytest.mark.validate
def test_get_effective_policy_summary_group_with_policy_assigned():
    ag_group_1 = get_asset_group("effective-policy-test-1")
    ag_group_2 = get_asset_group("effective-policy-test-2")

    response = get_effective_policy_summary(ag_group_1.get("id"))
    assert 3 == len(response["summaries"])

    total_assets = 0
    for summary in response["summaries"]:
        total_assets += summary["asset_count"]
    assert 5 == total_assets

    endpoint_1 = get_endpoint("pas-e2e-effective-1")
    endpoint_2 = get_endpoint("pas-e2e-effective-2")
    endpoint_3 = get_endpoint("pas-e2e-effective-3")
    endpoint_4 = get_endpoint("pas-e2e-effective-4")
    endpoint_5 = get_endpoint("pas-e2e-effective-5")

    assert contains_summary_with(response["summaries"], 1, POLICIES[3].get("id"), [ag_group_1.get("id")], "GROUP", [ endpoint_3.get("id") ])
    assert contains_summary_with(response["summaries"], 2, POLICIES[3].get("id"), [ag_group_1.get("id"), ag_group_2.get("id")], "GROUP", [ endpoint_1.get("id"), endpoint_2.get("id") ])
    assert contains_summary_with(response["summaries"], 2, POLICIES[-1].get("id"), [], "MANUAL", [ endpoint_4.get("id"), endpoint_5.get("id") ])


@pytest.mark.validate
def test_get_effective_policy_summary_group_with_no_members():
    ag_group_6 = get_asset_group("effective-policy-test-6")
    response = get_effective_policy_summary(ag_group_6.get("id"))
    assert 0 == len(response["summaries"])


def get_asset_group(name):
    ag = None
    for group in ASSET_GROUPS:
        if group["name"] == name:
            ag = group
            break

    assert ag is not None, f"{name} not found."
    return ag


def get_endpoint(name):
    e = None
    for endpoint in ENDPOINTS:
        if endpoint["name"] == name:
            e = endpoint
            break

    assert e is not None, f"endpoint {name} not found."
    return e


def contains_summary_with(records, asset_count, policy_id, asset_group_ids, policy_assignment_type, asset_ids):
    for record in records:
        if (
            record["asset_count"] == asset_count
            and record["policy_id"] == policy_id
            and [group.get("id") for group in record["asset_groups"]] == asset_group_ids
            and record["policy_assignment_type"] == policy_assignment_type
        ):
            if asset_ids is not None:
                affected_asset_ids = search_assets(SESSION, search_definition=record["assets_search_definition"])
                return set(affected_asset_ids) == set(asset_ids)
            else:
                return True

    return False


def get_effective_policy_summary(asset_group_id):
    response = SESSION.get(
        url=f"{LOCAL_PAS_HOST}/orgs/{SESSION.get_user_org_key()}/asset_groups/{asset_group_id}/policy_summary"
    )
    assert response.ok, EXPECTED_CALL_TO_SUCCEED
    return json.loads(response.content)


# to be used only when cleanup is performed on test org
def setup_effective_policy_test_data():
    # create endpoints
    endpoint_1 = sim_endpoint(SESSION, suffix="effective-1", device_type=SensorDeviceType.MAC)
    endpoint_2 = sim_endpoint(SESSION, suffix="effective-2", device_type=SensorDeviceType.MAC)
    endpoint_3 = sim_endpoint(SESSION, suffix="effective-3", device_type=SensorDeviceType.MAC)
    endpoint_4 = sim_endpoint(SESSION, suffix="effective-4", device_type=SensorDeviceType.MAC)
    endpoint_5 = sim_endpoint(SESSION, suffix="effective-5", device_type=SensorDeviceType.MAC)
    endpoint_6 = sim_endpoint(SESSION, suffix="effective-6", device_type=SensorDeviceType.MAC)

    # create asset groups
    group_1 = create_asset_group(SESSION, name="effective-policy-test-1", policy_id=POLICIES[3].get("id"))
    group_2 = create_asset_group(SESSION, name="effective-policy-test-2", policy_id=POLICIES[3].get("id"))
    group_3 = create_asset_group(SESSION, name="effective-policy-test-3", policy_id=POLICIES[4].get("id"))
    group_4 = create_asset_group(SESSION, name="effective-policy-test-4", policy_id=POLICIES[5].get("id"))
    group_5 = create_asset_group(SESSION, name="effective-policy-test-5")

    # assign asset groups to endpoints
    add_asset_group_to_endpoint(session=SESSION, asset_group_id=group_5.get("id"), endpoint_id=endpoint_1.get("device_id"))
    add_asset_group_to_endpoint(session=SESSION, asset_group_id=group_5.get("id"), endpoint_id=endpoint_2.get("device_id"))
    add_asset_group_to_endpoint(session=SESSION, asset_group_id=group_5.get("id"), endpoint_id=endpoint_3.get("device_id"))
    add_asset_group_to_endpoint(session=SESSION, asset_group_id=group_5.get("id"), endpoint_id=endpoint_4.get("device_id"))
    add_asset_group_to_endpoint(session=SESSION, asset_group_id=group_5.get("id"), endpoint_id=endpoint_5.get("device_id"))
    add_asset_group_to_endpoint(session=SESSION, asset_group_id=group_5.get("id"), endpoint_id=endpoint_6.get("device_id"))

    add_asset_group_to_endpoint(session=SESSION, asset_group_id=group_1.get("id"), endpoint_id=endpoint_1.get("device_id"))
    add_asset_group_to_endpoint(session=SESSION, asset_group_id=group_2.get("id"), endpoint_id=endpoint_1.get("device_id"))
    add_asset_group_to_endpoint(session=SESSION, asset_group_id=group_3.get("id"), endpoint_id=endpoint_1.get("device_id"))

    add_asset_group_to_endpoint(session=SESSION, asset_group_id=group_1.get("id"), endpoint_id=endpoint_2.get("device_id"))
    add_asset_group_to_endpoint(session=SESSION, asset_group_id=group_2.get("id"), endpoint_id=endpoint_2.get("device_id"))
    add_asset_group_to_endpoint(session=SESSION, asset_group_id=group_4.get("id"), endpoint_id=endpoint_2.get("device_id"))

    add_asset_group_to_endpoint(session=SESSION, asset_group_id=group_1.get("id"), endpoint_id=endpoint_3.get("device_id"))

    add_asset_group_to_endpoint(session=SESSION, asset_group_id=group_1.get("id"), endpoint_id=endpoint_4.get("device_id"))
    add_asset_group_to_endpoint(session=SESSION, asset_group_id=group_3.get("id"), endpoint_id=endpoint_4.get("device_id"))
    add_policy_override(session=SESSION, policy_id=POLICIES[-1].get("id"), device_id=endpoint_4.get("device_id"))

    add_asset_group_to_endpoint(session=SESSION, asset_group_id=group_1.get("id"), endpoint_id=endpoint_5.get("device_id"))
    add_policy_override(session=SESSION, policy_id=POLICIES[-1].get("id"), device_id=endpoint_5.get("device_id"))
